package com.werken.classworlds;

/*
 $Id: RealmClassLoader.java,v 1.6 2002/12/14 08:32:27 bob Exp $

 Copyright 2002 (C) The Werken Company. All Rights Reserved.
 
 Redistribution and use of this software and associated documentation
 ("Software"), with or without modification, are permitted provided
 that the following conditions are met:

 1. Redistributions of source code must retain copyright
    statements and notices.  Redistributions must also contain a
    copy of this document.
 
 2. Redistributions in binary form must reproduce the
    above copyright notice, this list of conditions and the
    following disclaimer in the documentation and/or other
    materials provided with the distribution.
 
 3. The name "classworlds" must not be used to endorse or promote
    products derived from this Software without prior written
    permission of The Werken Company.  For written permission,
    please contact bob@werken.com.
 
 4. Products derived from this Software may not be called "classworlds"
    nor may "classworlds" appear in their names without prior written
    permission of The Werken Company. "classworlds" is a registered
    trademark of The Werken Company.
 
 5. Due credit should be given to The Werken Company.
    (http://classworlds.werken.com/).
 
 THIS SOFTWARE IS PROVIDED BY THE WERKEN COMPANY AND CONTRIBUTORS
 ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 THE WERKEN COMPANY OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.
 
 */

import java.net.URL;
import java.security.SecureClassLoader;
import java.io.InputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.jar.JarInputStream;
import java.util.jar.JarEntry;

/** Classloader for <code>ClassRealm</code>s.
 *
 *  @author <a href="mailto:bob@eng.werken.com">bob mcwhirter</a>
 *
 *  @version $Id: RealmClassLoader.java,v 1.6 2002/12/14 08:32:27 bob Exp $
 */
class RealmClassLoader
    extends SecureClassLoader
{
    // ------------------------------------------------------------
    //     Constants
    // ------------------------------------------------------------
    
    /** Empty URL array. */
    private static final URL[] EMPTY_URL_ARRAY = new URL[0];

    // ------------------------------------------------------------
    //     Instance members
    // ------------------------------------------------------------

    /** The realm. */
    private ClassRealmImpl realm;

    /** Urls. */
    private List urls;

    /** Classes, indexed by path. */
    private Map classIndex;

    // ------------------------------------------------------------
    //     Constructors
    // ------------------------------------------------------------

    /** Construct.
     *
     *  @param realm The realm for which this loads.
     */
    RealmClassLoader(ClassRealmImpl realm)
    {
        super( null );

        this.realm      = realm;
        this.urls       = new ArrayList();
        this.classIndex = new HashMap();
    }

    // ------------------------------------------------------------
    //     Instance methods
    // ------------------------------------------------------------

    /** Retrieve the realm.
     *
     *  @return The realm.
     */
    ClassRealmImpl getRealm()
    {
        return this.realm;
    }

    /** Add a constituent to this realm for locating classes.
     *
     *  @param constituent URL to contituent jar or directory.
     */
    void addConstituent(URL constituent)
    {
        this.urls.add( constituent );
    }

    /** Load a class directly from this classloader without
     *  defering through any other <code>ClassRealm</code>.
     *
     *  @param name The name of the class to load.
     *
     *  @return The loaded class. 
     *
     *  @throws ClassNotFoundException If the class could not be found.
     */
    Class loadClassDirect(String name) throws ClassNotFoundException
    {
        return super.loadClass( name,
                                true );
    }

    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    //     java.lang.ClassLoader
    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

    /** @see ClassLoader
     */
    public URL findResource(String name)
    {
        URL resourceUrl = null;

        Iterator urlIter = this.urls.iterator();
        URL      eachUrl = null;

        while ( urlIter.hasNext() )
        {
            eachUrl = (URL) urlIter.next();

            try
            {
                resourceUrl = new URL( eachUrl,
                                       name );
                
                InputStream in = resourceUrl.openStream();

                in.close();

                return new URL( eachUrl,
                                name );
            }
            catch (IOException e)
            {
                // swallow
            }
        }

        return null;
    }

    /** @see ClassLoader
     */
    public synchronized Class findClass(String className)
        throws ClassNotFoundException
    {
        String classPath = className.replace( '.',
                                              '/' ) + ".class";

        if ( this.classIndex.containsKey( classPath ) )
        {
            return (Class) this.classIndex.get( classPath );
        }

        Iterator urlIter = this.urls.iterator();
        URL      eachUrl = null;

        byte[]   classBytes = null;

        while ( ( classBytes == null )
                &&
                ( urlIter.hasNext() ) )
        {
            eachUrl = (URL) urlIter.next();
            
            if ( "jar".equals( eachUrl.getProtocol() )
                 ||
                 eachUrl.toExternalForm().endsWith( ".jar" ) )
            {
                classBytes = findClassInJarStream( eachUrl,
                                                   classPath );
            }
            else
            {
                classBytes = findClassInDirectoryUrl( eachUrl,
                                                      classPath );
            }
        }

        if ( classBytes == null )
        {
            throw new ClassNotFoundException( className );
        }

        Class cls = defineClass( className,
                                 classBytes,
                                 0,
                                 classBytes.length );

        this.classIndex.put( className,
                             cls );

        return cls;
    }

    /** Attempt to load the bytes of a class from a JAR <code>URL</code>.
     *
     *  @param inUrl The base url.
     *  @param path The path to the desired class.
     *
     *  @return The class bytes or <code>null</code> if not available
     *          via the base url.
     */
    protected byte[] findClassInJarStream(URL inUrl,
                                          String path)
    {
        try
        {
            JarInputStream in = new JarInputStream( inUrl.openStream() );
            
            try
            {
                JarEntry entry = null;
                
                while ( in.available() != 0 )
                {
                    entry = in.getNextJarEntry();

                    if ( entry == null )
                    {
                        return null;
                    }

                    if ( entry.getName().equals( path ) )
                    {
                        ByteArrayOutputStream out = new ByteArrayOutputStream();
                        
                        try
                        {
                            byte[] buffer = new byte[ 2048 ];
                            
                            int read = 0;
                            
                            while ( in.available() > 0 )
                            {
                                read = in.read( buffer,
                                                0,
                                                buffer.length );
                                
                                if ( read < 0 )
                                {
                                    break;
                                }
                                
                                out.write( buffer,
                                           0,
                                           read );
                            }
                            
                            return out.toByteArray();
                        }
                        finally
                        {
                            out.close();
                        }
                    }
                }
            }
            finally
            {
                in.close();
            }
        }
        catch (IOException e)
        {
            // swallow
        }

        return null;
    }
    
    /** Attempt to load the bytes of a class from a directory <code>URL</code>.
     *
     *  @param url The directory url.
     *  @param path The path tot he desired class.
     *
     *  @return The class bytes or <code>null</code> if not available
     *          via the base url.
     */
    protected byte[] findClassInDirectoryUrl(URL url,
                                             String path)
    {
        try
        {
            URL classUrl = new URL( url,
                                    path );
        }
        catch (IOException e)
        {
            // swallow
        }
            
        return null;
    }

    /** Load a class.
     *
     *  @param name The name of the class to load.
     *  @param resolve If <code>true</code> then resolve the class.
     *
     *  @return The loaded class. 
     *
     *  @throws ClassNotFoundException If the class cannot be found.
     */
    protected Class loadClass(String name,
                              boolean resolve) throws ClassNotFoundException
    {
        return getRealm().loadClass( name );
    }

    /** Retrieve the <code>URL</code>s used by this <code>ClassLoader</code>.
     *
     *  @return The urls.
     */
    public URL[] getURLs()
    {
        return (URL[]) this.urls.toArray( EMPTY_URL_ARRAY );
    }
}
